#include <Stepper.h>
#include <menu.h>
#include <menuIO/serialIO.h>
#include <menuIO/stringIn.h>
#include <menuIO/chainStream.h>
#include <menuIO/liquidCrystalOut.h>
#include <EEPROM.h>
#include <LiquidCrystal.h>
#include "decimalField.h"
using namespace Menu;

//Definice pinu desky
#define RESET A0
#define BACK_BTN A1
#define SEL_BTN 2
#define NAV_BTN 3

#define MAX_DEPTH 2
#define SOFT_DEBOUNCE_MS 100

#define INT_MAX 32767
//Adresy v pameti
#define STP_ADR 1
#define SPD_ADR 3
#define MOD_ADR 7
#define TOPSPD_ADR 9
#define MINSPD_ADR 13
#define INIT_ADR 17
#define VOL_ADR 21

const char * calEnd = "Calibration ended.";
const char * pumped = "Pumped: ";

//pripojeni cerpadla
#define MOT1 10
#define MOT2 11
#define MOT3 12
#define MOT4 13
//konfigurace motoru cerpadla
#define MOTSPDL 1
#define MOTSPDH 1280 // 20*64  20 = rpm čerpadla po zpřevodování motoru 1:64
#define MOTSTEPS 200 // uhel kroku čerpadla po zprevodování je 0,0282, 0,0282*64 = 1.8 uhel kroku motoru => 360/1,8 = 200 kroků

char INIT = EEPROM.read(INIT_ADR);
//Nacteni kalibracnich hodnot
float topSpeed = EEPROM.get(TOPSPD_ADR,topSpeed); 
float minSpeed = EEPROM.get(MINSPD_ADR,minSpeed);
uint16_t stepsPerDeciMl = EEPROM.read(STP_ADR) + EEPROM.read(STP_ADR+1) * 256;  

unsigned long endTime = 0;
//definice promenych pro menu
char mode, defMode = 1;
float spd = minSpeed, spdDef = minSpeed;
float volf = 1.0,volCalf = 1.0, volDeff = 1.0;

bool stopPump = false;

void stopPumping() { //funkce pro obsluhu preruseni
  stopPump = true;
  endTime = millis();
}

void initSets() { //Nastavení hodnot při prvním spuštění na konkrétní desce nebo po smazání EEPROM paměti
  EEPROM.put(VOL_ADR,10.0);
  EEPROM.update(STP_ADR,1);
  EEPROM.update(STP_ADR+1,0);
  EEPROM.put(SPD_ADR, 1.0);
  EEPROM.update(MOD_ADR, 1.0);
  EEPROM.put(TOPSPD_ADR, 10.0);
  EEPROM.put(MINSPD_ADR, 1.0);
  EEPROM.update(INIT_ADR, 1);
  digitalWrite(RESET, LOW);
}



//Definice objektů pro LCD displej a motor
LiquidCrystal lcd(4, 5, 6, 7, 8, 9);
Stepper motor = Stepper(MOTSTEPS, MOT4, MOT3, MOT2, MOT1); //predefinovat s opravym motorem

void cPump(float pumpSpeed, bool serial = false){
      float currentVolume = 0.0;
      uint16_t motorSpeed = map(pumpSpeed, minSpeed, topSpeed, MOTSPDL, MOTSPDH);
      motor.setSpeed(motorSpeed);
   do {
      motor.step(stepsPerDeciMl);
      currentVolume += (0.1); 
      if(serial){
        Serial.print(pumped);
        Serial.print(currentVolume,2);Serial.print("ml");
        Serial.println();
        } 
      else{
        lcd.clear();
        lcd.print(pumped);
        lcd.print(currentVolume,2);lcd.print("ml      ");
        }

      if(Serial.available()){
        char c;
        c = Serial.read();
        if(c==10 || c==13 || c==64){
          stopPump = true;
        }
      }
   }while (!stopPump);
}

void sPump(float pumpSpeed, float pumpVolume, bool serial = false){
  float currentVolume = 0.0;
  uint16_t motorSpeed = map((uint16_t)(pumpSpeed*10), (uint16_t)(minSpeed*10), (uint16_t)(topSpeed*10), MOTSPDL, MOTSPDH);
  motor.setSpeed(motorSpeed);
  while (!stopPump) {
      lcd.setCursor(0, 0);
      motor.step(stepsPerDeciMl);
      currentVolume += 0.1;
      if(serial){
        Serial.println(pumped);
        Serial.print(currentVolume,2); Serial.print("/");Serial.print(pumpVolume,1); Serial.print("ml");Serial.println("");
        } 
      else{
        lcd.print(pumped);
        lcd.print(currentVolume,2); lcd.print("/"); lcd.print(pumpVolume,1); lcd.print("ml");
        }
      if (currentVolume >= pumpVolume) {
        stopPump = true;
      }
   }
}

result runPump() {
  attachInterrupt(0, stopPumping, RISING);
  lcd.setCursor(0, 1);
  lcd.print("Stop      ");
  stopPump = false;
  if (mode == 1)
    cPump(spd);
  else
    sPump(spd, volf);
  lcd.setCursor(0, 0);
  lcd.print("Done!  ");
  lcd.setCursor(0, 1);
  lcd.print("Run again");
  detachInterrupt(0);
  return proceed;

}



unsigned long calibPump(int motorSpeed, unsigned long * steps, bool serial=false) {
  while (!digitalRead(SEL_BTN)) {}
  attachInterrupt(0, stopPumping, RISING);
  motor.setSpeed(motorSpeed);
  stopPump = false;
  unsigned long currentTime, startTime = 0;
  startTime = millis();
  * steps = 0;
  uint16_t stepsPre = 0;
  if(serial){
    Serial.println("Press enter at end");
    }
  do {
   currentTime = (millis() - startTime) / 1000;
   if(serial){
    char c;
      Serial.print("Time: ");
      Serial.print(currentTime);
      Serial.print(" s");
      Serial.print(" | ");
      Serial.print(* steps);
      Serial.println();
      if(Serial.available()){
      c = Serial.read();
      if(c==10 || c==13 || c==64){
        stopPump = true;
        endTime=(currentTime*1000)+startTime;
        }
      }
   }
   else{
    lcd.clear();
    lcd.print("Time: ");
    lcd.print(currentTime);
    lcd.print(" s");
    lcd.print(" | ");
    lcd.print(* steps);
    lcd.setCursor(0, 1);
    lcd.print(">Stop(Vol. reached)");
   }
    motor.step(1); * steps += 1;
   
  } while (!stopPump);
  detachInterrupt(0);
  return endTime - startTime;
}

void runCalib(){
  runCalibration(false);
  }

void runCalibration(bool serial) {
  if(serial){
    Serial.println("Step 1 - low speed ");
    Serial.println("Press Enter ...");
    char c;
    while(c!=10 && c!=13 && c!=64){
      if(Serial.available()){
      c = Serial.read();
      }
    }
  }
  else{
    lcd.clear();
    lcd.print("Step 1 - low speed "); 
    lcd.setCursor(0, 1);
    lcd.print(">Start calib. step.");
    delay(SOFT_DEBOUNCE_MS);
    while (digitalRead(SEL_BTN)) {
    delay(SOFT_DEBOUNCE_MS);
    } 
  }
  unsigned long stepsL = 0;
  unsigned long calibTimeL = calibPump(MOTSPDL, &stepsL,serial);

  if(serial){
    Serial.println("Step 2 - high speed");
    Serial.println("Press Enter ...");
    char c;
    while(c!=10 && c!=13 && c!=64){
      if(Serial.available()){
      c = Serial.read();
      }
    }     
   }
  else{
    lcd.clear();
    lcd.print("Step 2 - high speed");  //tady bude krok motoru
    lcd.setCursor(0, 1);
    lcd.print(">Start calib. step.");
    delay(SOFT_DEBOUNCE_MS);
    while (digitalRead(SEL_BTN)) {
      delay(SOFT_DEBOUNCE_MS);
    }
  }
  
  unsigned long stepsH = 0;
  unsigned long calibTimeH = calibPump(MOTSPDH, &stepsH,serial);
  
  topSpeed = volCalf / ((float)calibTimeH / 60000.0);
  minSpeed = volCalf / ((float)calibTimeL / 60000.0);
  bool failed = false;
  if(minSpeed>=topSpeed ){
    if(serial){
      Serial.println("Calibration failed.");
    }
    else{
      lcd.clear();
      lcd.print("Calibration failed.");
      lcd.setCursor(0, 1);
      }
      minSpeed = 1.0;
      topSpeed = 10.0;
      failed = true;
    }
  else if(minSpeed == 0.0){
      minSpeed = 1.0;
  
  }

  uint16_t volDeciML = (uint16_t)(volCalf * 10);
  unsigned long stepsAvr = (stepsL+stepsH)/2;
  stepsPerDeciMl = stepsAvr / (unsigned long)(volCalf * 10);
  
if(serial){
    Serial.println(calEnd);
    Serial.println("Restart in: ");
    for (int i = 5; i > 0; i--) {
    Serial.print(i);
    Serial.print(" s");
    Serial.println();
    delay(1000);
  }
  }
  else{
    
    if(!failed){
    lcd.clear();
    lcd.print(calEnd);
      }
    
    lcd.setCursor(0, 1);
    lcd.print("Restart in: ");
    for (int i = 5; i > 0; i--) {
     lcd.print(i);
     lcd.print(" s");
     lcd.setCursor(12, 1);
     delay(1000);

    }
  } 
  

  EEPROM.put(TOPSPD_ADR, topSpeed);
  EEPROM.put(MINSPD_ADR, minSpeed);
  EEPROM.update(STP_ADR, stepsPerDeciMl % 256);
  EEPROM.update(STP_ADR + 1, stepsPerDeciMl / 256);

  digitalWrite(RESET, LOW);

}

int doSerialCommand(char * line){
   //return value:
   // -3  Invalid parameters
   // -2: Invalid syntax
   // -1: se comand
   //  1: Exit
   //  0: Command execude
  char * space = strchr(line, ' ');
  if(strncmp(line,"info",strlen(line))==0 && space == NULL){
    Serial.println("calibrate VOL");
    Serial.println("pump -c SPD");
    Serial.println("pump -s VOL SPD");
    Serial.println("settings VOL SPD MOD");
    Serial.println("exit");
    return 0;
  }
  else if(strncmp(line,"exit",strlen(line))==0 && space==NULL){
    //konec
    return 1;
    }
   else if (space==NULL){
    return -1;
    }
   
  char * command = line;
  * space = '\0';
  char * param;
  int len = 0;
  float spd;
  float volf;
  int spc;
  if(strncmp(command,"pump",strlen(command))==0){
    char * dash = strchr(space+1, '-');
    if(dash==NULL){
      return -2;
      //invalid syntax
      }
    else{
      param = dash+1;
      }
    if(* param=='c' || * param=='C'){
          
          spd = atof(dash+3);
          if(spd == 0.0 || spd>topSpeed || spd<minSpeed){

            return -3;
            }
            attachInterrupt(0, stopPumping, RISING);
            stopPump = false;
            cPump(spd,true);
            Serial.println("Done!  ");
                
     }else if (* param=='s' || * param=='S'){
          char * space2 = strchr(param, ' ');
          
          if(space2 == NULL){
            return -2;
            }
           * space2 = '\0';
          spd = atof(dash+3);
          volf = atof(space2+1);
          if(spd == 0.0 || volf==0.0 || spd<minSpeed || spd>topSpeed || volf<0){
            //invalid syntax 
            return -3;
            }    
            attachInterrupt(0, stopPumping, RISING);
            stopPump = false;
            sPump(spd,volf,true); 
            Serial.println("Done!  "); 
      }else{
        return -2;
        }
  
  }else if (strncmp(command,"calibrate",strlen(command))==0){
      volf = atof(space+1);
      if(volf==0.0 || volf<0){ return -3;}
      runCalibration(true);
      
  }else if(strncmp(command,"settings",strlen(command))==0){
    char * space2 = strchr(space+1, ' ');
    if (space2 == NULL)
      return -2;
    char * space3 = strchr(space2+1, ' ');
    if (space3 == NULL)
      return -2;
    * space2 = '\0'; * space3 = '\0';
    
    float vols = atof(space+1);
    float spds = atof(space2+1);
    int mods = atoi(space3+1); 
    if(vols==0.0 || spds==0.0 || mods == 0 || spds<minSpeed || spds>topSpeed || volf<0){
      Serial.println(vols);
      Serial.println(spds);
      Serial.println(mods);
      return -3;
      }
    else{
      volDeff = vols;
      spdDef = spds;
      defMode = mods;
      saveSet();
      }
   }else{
     //invalid command  
     return -1;
    
  }
  return 0;
  }

void serialRemote() { //https://www.root.cz/clanky/ovladani-arduina-v-realnem-case-z-pocitace/
  
  Serial.begin(9600);
  Serial.println ("Pump Driver");
  Serial.print(">");
  char lineCmd [60];
  int i=0;
  char c;
  int res;
  while(true){
   if(Serial.available()){
    c = Serial.read(); 
    
    Serial.print(c);
     
    if(c==10 || c==13 || c==64){//newline
      Serial.print("\n");
        if(i>60){res = -2;}
        else{
          res = doSerialCommand(lineCmd);
        }
        switch(res){
          case -2:{
             Serial.println("Invalid syntax");
             Serial.print(">");
             break;
          }
          case -1:{
             Serial.println("Invalid command");
             Serial.print(">");
              break;
          }
          case -3:{
              Serial.println("Invalid parameters");
              Serial.print(">");
              break;
          }
          case 1:{
             Serial.println("Good bye!");
             break;
          }
          case 0:{
            Serial.print(">");  
          }           
        }  
      lineCmd[0]='\0'; i=0;
      
      if(res==1){
        break;
      } 
    }
    else if(c==8 || c==127){
      i--;
      if(i<0)
        i=0;
      lineCmd[i]='\0';
      
      }
    else{
      lineCmd[i]=c;
      lineCmd[i+1]='\0';
      i++;
    } 
   }  
  }                                                                    
  Serial.end();
  
  }
  
void saveSet() {

  EEPROM.put(VOL_ADR,volDeff);
  EEPROM.put(SPD_ADR, spdDef);
  EEPROM.update(MOD_ADR, defMode);

  volf = volDeff;
  spd = spdDef;
  mode = defMode;
  lcd.setCursor(0,1);
  lcd.print("Saved");
  Serial.println("Saved");
}

void readSet() {

  EEPROM.get(VOL_ADR, volDeff);
  EEPROM.get(SPD_ADR,spdDef);
  defMode = EEPROM.read(MOD_ADR);
  if (defMode != 2 && defMode != 1) {
    defMode = 1; 
    EEPROM.update(MOD_ADR, defMode);
  }
  if (spdDef > topSpeed) {
    spdDef = topSpeed;
  }
  if(spdDef < minSpeed){
    spdDef = minSpeed;
    }
  if (volDeff > 500.0)
  {
    volDeff = 500.0;
  }
}

// https://github.com/neu-rah/ArduinoMenu/blob/a0440c39a9de5453c1b2141035e6b6eb11e1fd1e/examples/dynamic/dynamic/dynamic.ino

TOGGLE(mode, modeMenu, "Mode: ", doNothing, noEvent, wrapStyle
       , VALUE("Continuous", 1, doNothing, noEvent)
       , VALUE("Specific volume", 2, doNothing, noEvent)
      );
TOGGLE(defMode, modeMenuDef, "Def. mode: ", doNothing, noEvent, wrapStyle
       , VALUE("Continuous", 1, doNothing, noEvent)
       , VALUE("Specific volume", 2, doNothing, noEvent)
      );

prompt* startMenuData[] = {
  &modeMenu,
  new decimalslField<typeof(volf)>(volf, "Volume","ml",1.0,500.0,0.1,1,doNothing,noEvent,wrapStyle),
  new decimalslField<typeof(spd)>(spd, "Speed","ml/min",minSpeed,topSpeed,0.1,1,doNothing,noEvent,wrapStyle),
  new prompt("Run", (Menu::callback)runPump, enterEvent ),
};

menuNode& startMenu = *new menuNode(
                        "Start",
                        sizeof(startMenuData) / sizeof(prompt*),
                        startMenuData,
                        doNothing,
                        noEvent,
                        wrapStyle
                      );

prompt* calibMenuData[] = {
  new decimalslField<typeof(volCalf)>(volCalf, "Volume","ml",1.0,500.0,0.1,1,doNothing,noEvent,wrapStyle),
  new prompt("Run", (Menu::callback)runCalib, enterEvent),
};
menuNode& calibMenu = *new menuNode(
                        "Calibration",
                        sizeof(calibMenuData) / sizeof(prompt*),
                        calibMenuData,
                        doNothing,
                        noEvent,
                        wrapStyle);

prompt* setMenuData[] = {
  &modeMenuDef,
  new decimalslField<typeof(volDeff)>(volDeff, "Def.vol","ml",1.0,500.0,0.1,1,doNothing,noEvent,wrapStyle),
  new decimalslField<typeof(spdDef)>(spdDef, "Def.spd","ml/min",minSpeed,topSpeed,0.1,1,doNothing,noEvent,wrapStyle),
  new prompt("Save", (Menu::callback)saveSet, enterEvent),

};
menuNode& setMenu = *new menuNode(
                      "Settings",
                      sizeof(setMenuData) / sizeof(prompt*),
                      setMenuData,
                      doNothing,
                      noEvent,
                      wrapStyle
                    );

prompt* pcCtrlMenuData[] = {
  new prompt("Connect", (Menu::callback)serialRemote, enterEvent),
};
menuNode& pcCtrlMenu = *new menuNode(
                         "PC Control",
                         sizeof(pcCtrlMenuData) / sizeof(prompt*),
                         pcCtrlMenuData,
                         doNothing,
                         noEvent,
                         wrapStyle
                       );

prompt* topMenuData[] = {
  &startMenu,
  &calibMenu,
  &setMenu,
  &pcCtrlMenu,
  
};
menuNode& topMenu = *new menuNode(
                      "Pump Driver",
                      sizeof(topMenuData) / sizeof(prompt*),
                      topMenuData,
                      doNothing,
                      noEvent,
                      wrapStyle
                    );



MENU_OUTPUTS(out, MAX_DEPTH
             , NONE
             , LIQUIDCRYSTAL_OUT(lcd, {0, 0, 16, 2}) //must have 2 items at least
            );

//MENU_INPUTS(in);
chainStream<0> in(NULL);
NAVROOT(nav, topMenu, MAX_DEPTH, in, out);




void setup() {
  interrupts(); // zapnutí podpory přerušení
  digitalWrite(RESET, HIGH); 
  delay(200);
  pinMode(RESET, OUTPUT);
  if (INIT != 1) {
    initSets(); // inicializace při prvním spuštění aplikace
  }
  //konfigurace vstupních avýstupních pinů
  pinMode(NAV_BTN, INPUT_PULLUP);
  pinMode(SEL_BTN, INPUT_PULLUP);
  pinMode(BACK_BTN, INPUT_PULLUP);

  pinMode(MOT1, OUTPUT);
  pinMode(MOT2, OUTPUT);
  pinMode(MOT3, OUTPUT);
  pinMode(MOT4, OUTPUT);

  lcd.begin(16, 2);
 
  readSet();
  volf = volDeff;
  spd = spdDef;
  mode = defMode;

}

void loop() {
  
  if (!digitalRead(SEL_BTN)) {
    delay(SOFT_DEBOUNCE_MS);
    while (!digitalRead(SEL_BTN));
    nav.doNav(enterCmd);
    delay(SOFT_DEBOUNCE_MS);
  }
  if (!digitalRead(NAV_BTN)) {
    delay(SOFT_DEBOUNCE_MS);
    while (!digitalRead(NAV_BTN)); //wait for button release
    nav.doNav(upCmd);
    delay(SOFT_DEBOUNCE_MS);
  }
  if (!digitalRead(BACK_BTN) && &(nav.active()) != &topMenu) {
    delay(SOFT_DEBOUNCE_MS);
    while (!digitalRead(BACK_BTN)); //wait for button release
    nav.doNav(escCmd);
    delay(SOFT_DEBOUNCE_MS);
  }
  nav.doOutput();
}
